package binarytree.core;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

import org.junit.Test;

public class BinarySearchTreeTest {
	public final static String CHARS = "hdlbfjn";

	BinarySearchTree<String> tree = new BinarySearchTree<>();

	public void foreach(String chars, Consumer<String> innerLoop) {
		for (int i = 0; i < chars.length(); i++) {
			innerLoop.accept(chars.substring(i, i+1));
		}
	}

	public void foreachReversed(String chars, Consumer<String> innerLoop) {
		for (int i = chars.length() - 1; i >= 0; i--) {
			innerLoop.accept(chars.substring(i, i+1));
		}
	}

	public void foreachIndexed(String chars, BiConsumer<Integer,String> innerLoop) {
		for (int i = 0; i < chars.length(); i++) {
			innerLoop.accept(i, chars.substring(i, i+1));
		}
	}

	public static <T extends Comparable<T>> boolean isSorted(ArrayList<T> list) {
		for (int i = 0; i < list.size() - 1; i++) {
			if (list.get(i).compareTo(list.get(i+1)) > 0) {
				return false;
			}
		}
		return true;
	}

	@Test
	public void testLevelOrder() {
		test1();
		ArrayList<ArrayList<String>> levels = tree.levelorder();
		assertEquals(3, levels.size());
		assertEquals(1, levels.get(0).size());
		assertEquals(2, levels.get(1).size());
		assertEquals(4, levels.get(2).size());

		// Nodes
		assertEquals("h", levels.get(0).get(0));
		assertEquals("d", levels.get(1).get(0));
		assertEquals("l", levels.get(1).get(1));
		assertEquals("b", levels.get(2).get(0));
		assertEquals("f", levels.get(2).get(1));
		assertEquals("j", levels.get(2).get(2));
		assertEquals("n", levels.get(2).get(3));
	}

	@Test
	public void test1() {
		foreach(CHARS, c -> tree.insert(c));
		foreach(CHARS, c -> assertTrue(tree.contains(c)));
		foreach("ABCDEFGHIJKLMNOPQRSTUVWXYZpqrstuvwxyz", c -> assertFalse(tree.contains(c)));
	}

	@Test
	public void testSize() {
		foreach(CHARS, c -> tree.insert(c));
		assertEquals(CHARS.length(), tree.size());
	}

	@Test
	public void testInorder() {
		test1();
		assertTrue(isSorted(tree.inorder()));
	}

	@Test
	public void testPreorder() {
		test1();
		ArrayList<String> pre = tree.preorder();
		foreachIndexed("hdbfljn", (i, c) -> assertEquals(c, pre.get(i)));
	}

	@Test
	public void testPostorder() {
		test1();
		ArrayList<String> post = tree.postorder();
		foreachIndexed("bfdjnlh", (i, c) -> assertEquals(c, post.get(i)));
	}

	@Test
	public void testRemove1() {
		test1();
		foreachReversed(CHARS, c -> assertRemove(c));
	}

	@Test
	public void testRemove2() {
		test1();
		foreach(CHARS, c -> assertRemove(c));
	}

	@Test
	public void testRemove3() {
		test1();
		foreach("njfblhd", c -> assertRemove(c));
	}

	@Test
	public void testRemove4() {
		test1();
		foreach("dbfjlnh", c -> assertRemove(c));
	}

	@Test
	public void testRemove5() {
		test1();
		foreach("hh", c -> assertRemove(c));
	}

	public void assertRemove(String c) {
		tree.remove(c);
		assertFalse(tree.contains(c));
	}
}
