package binarytree.core;

import java.util.ArrayList;

/**
 * A basic binary tree class.
 * 
 * @author CSCI 151
 *
 * @param <E>
 *            The type of data to be stored in nodes of the tree.
 */
public class BinaryTree<E> {

	// Is this the empty tree?
	private boolean isEmpty;

	// References to child trees.
	private BinaryTree<E> left, right;

	// The data stored in this node.
	private E data;

	/**
	 * Construct an empty binary tree.
	 */
	public BinaryTree() {
		isEmpty = true;
	}

	/**
	 * Construct a non-empty binary tree.
	 * 
	 * @param value
	 *            The value to store in the root.
	 * @param left
	 *            The left subtree.
	 * @param right
	 *            The right subtree.
	 */
	public BinaryTree(E value, BinaryTree<E> left, BinaryTree<E> right) {
		isEmpty = false;
		this.data = value;
		setLeft(left);
		setRight(right);
	}

	/**
	 * Construct a leaf.
	 * 
	 * @param value
	 *            The value to store in the leaf.
	 */
	public BinaryTree(E value) {
		this(value, new BinaryTree<E>(), new BinaryTree<E>());
	}

	public BinaryTree<E> getLeft() {
		return left;
	}

	public void setLeft(BinaryTree<E> left) {
		this.left = left;
	}

	public BinaryTree<E> getRight() {
		return right;
	}

	public void setRight(BinaryTree<E> right) {
		this.right = right;
	}

	public E getData() {
		return data;
	}

	/**
	 * Set the data at the root. If the tree was empty before, it is now non-empty
	 * and has new empty trees as its left and right children.
	 * 
	 * @param data
	 *            The data to store at the root.
	 */
	public void setData(E data) {
		this.data = data;
		if (isEmpty()) {
			isEmpty = false;
			setLeft(new BinaryTree<>());
			setRight(new BinaryTree<>());
		}
	}

	public boolean isEmpty() {
		return isEmpty;
	}

	/**
	 * Compute the total number of data values stored in the tree.
	 * 
	 * @return The count of data values.
	 */
	public int size() {
		if (isEmpty()) {
			return 0;
		} else {
			// Not empty, there is a value + left + right
			// assume left.size() and right.size() correctly
			// compute sizes of left & right subtrees
			return 1 + left.size() + right.size();
		}
	}

	@Override
	public String toString() {
		if (isEmpty()) {
			return "";
		} else if (isLeaf()) {
			return data.toString();
		} else {
			return "(" + left.toString() + ")" + data + "(" + right.toString() + ")";
		}
	}

	////////////////////////////////////////////////////////////

	/**
	 * Test whether the tree is non-empty and has a non-empty left child.
	 * 
	 * @return
	 */
	public boolean hasLeft() {
		// TODO Implement me
		return false;
	}

	/**
	 * Test whether the tree is non-empty and has a non-empty right child.
	 */
	public boolean hasRight() {
		// TODO Implement me
		return false;
	}

	/**
	 * Test whether the tree is a leaf (i.e. non-empty but both children are empty).
	 */
	public boolean isLeaf() {
		// TODO Implement me
		return false;
	}

	/**
	 * Get a reference to the subtree whose root is the rightmost node in the tree.
	 */
	public BinaryTree<E> rightmost() {
		// TODO Implement me
		return null;
	}

	/**
	 * Compute the height of the tree, which is the length of the longest path from
	 * the root to any leaf. The height of the empty tree is -1.
	 * 
	 * @return The height.
	 */
	public int height() {
		// TODO Implement me
		return 0;
	}

	/**
	 * Return a list containing an in-order traversal of the data.
	 */
	public ArrayList<E> inorder() {
		// Create a new ArrayList in which to put the elements
		ArrayList<E> elts = new ArrayList<E>();

		// Call a recursive helper method to fill in the elements
		inorderFill(elts);

		// Return it
		return elts;
	}

	/**
	 * Given an ArrayList, add all the elements from the tree to it using an inorder
	 * traversal. Note there may already be elements in the ArrayList; inorderFill
	 * should add new elements to the end.
	 * 
	 * @param elts
	 */
	private void inorderFill(ArrayList<E> elts) {
		// TODO Implement me
	}

	/**
	 * Return a list containing an preorder traversal of the data.
	 */
	public ArrayList<E> preorder() {
		// TODO Implement me
		return null;
	}

	/**
	 * Return a list containing an postorder traversal of the data.
	 */
	public ArrayList<E> postorder() {
		// TODO Implement me
		return null;
	}

	//////////////////////////////////////////////////////

	// Pay no attention to the man behind the curtain
	public ArrayList<ArrayList<E>> levelorder() {
		ArrayList<ArrayList<E>> levels = new ArrayList<>();

		ArrayList<BinaryTree<E>> curLevel = new ArrayList<BinaryTree<E>>();
		if (!isEmpty())
			curLevel.add(this);

		while (!curLevel.isEmpty()) {

			ArrayList<E> curLevelValues = new ArrayList<>();
			for (BinaryTree<E> t : curLevel)
				curLevelValues.add(t.getData());
			levels.add(curLevelValues);

			ArrayList<BinaryTree<E>> nextLevel = new ArrayList<BinaryTree<E>>();
			for (BinaryTree<E> t : curLevel) {
				if (t.hasLeft())
					nextLevel.add(t.getLeft());
				if (t.hasRight())
					nextLevel.add(t.getRight());
			}
			curLevel = nextLevel;
		}
		return levels;
	}

}